from typing import Literal, Union, Annotated
import os

from pydantic import BaseModel, Field
import boto3
import psycopg2


class EnvVar(BaseModel):
    """The name of the env var where an output value can be found"""

    name: str

    def get_value(self):
        return os.environ.get(self.name)


class AllowWrite(BaseModel):
    account_id: str
    role_name: str


class BucketOptions(BaseModel):
    bucket_name: str | None
    policies: list[AllowWrite] | None
    # todo: add more


class BucketOutput(BaseModel):
    pass


class Bucket(BaseModel):
    type: Literal["shuttle_aws.s3.Bucket"] = "shuttle_aws.s3.Bucket"
    options: BucketOptions
    output: BucketOutput | None = None

    def init(self):
        # print(f"Bucket initializing at runtime with {self.options=}")
        pass

    def is_cf_resource(self, changeset: dict) -> bool:
        return (
            changeset["ResourceType"] == "AWS::S3::Bucket"
            and changeset["PhysicalResourceId"] == self.options.bucket_name
        )

    def get_client(self):
        s3 = boto3.client("s3")
        # check that the bucket exists and we have access to it
        _response = s3.head_bucket(Bucket=self.options.bucket_name)
        return s3


class RdsPostgresOptions(BaseModel):
    # todo: fields: engine_version, db_name
    pass


class RdsPostgresOutput(BaseModel):
    engine: EnvVar
    username: EnvVar
    password: EnvVar
    host: EnvVar
    port: EnvVar
    dbname: EnvVar
    dbInstanceIdentifier: EnvVar


class RdsPostgres(BaseModel):
    type: Literal["shuttle_aws.rds.Postgres"] = "shuttle_aws.rds.Postgres"
    options: RdsPostgresOptions
    output: RdsPostgresOutput | None = None

    def init(self):
        # print(f"RDS initializing at runtime with {self.options=}")
        pass

    def is_cf_resource(self, changeset: dict) -> bool:
        return changeset["ResourceType"] == "AWS::RDS::DBInstance"

    def get_connection(self):
        assert isinstance(self.output, RdsPostgresOutput)
        username = self.output.username.get_value()
        password = self.output.password.get_value()
        host = self.output.host.get_value()
        port = self.output.port.get_value()
        dbname = self.output.dbname.get_value()
        return psycopg2.connect(
            f"postgres://{username}:{password}@{host}:{port}/{dbname}"
        )


class EcsCronTaskOptions(BaseModel):
    schedule: str


class EcsCronTask(BaseModel):
    type: Literal["shuttle.ecs.cron_task"] = "shuttle.ecs.cron_task"
    options: EcsCronTaskOptions


AllResourcesUnion = Union[Bucket, RdsPostgres]


class InfraManifest(BaseModel):
    # todo: replace with union model when there are multiple options
    service: EcsCronTask
    resources: (
        list[Annotated[AllResourcesUnion, Field(..., discriminator="type")]] | None
    )



import asyncio
import importlib.util
import json
import logging
import os
import sys
from pathlib import Path
from typing import Annotated

from shuttle_common import InfraManifest
from shuttle_infra import (
    inspect_stack,
)
from shuttle_infra.common import BotoSession

logger = logging.getLogger(__name__)


class ImportSpecError(Exception):
    pass


class ModuleExecError(Exception):
    pass


class RDSDescriptionError(Exception):
    """Exception raised when an RDS instance cannot be described."""


class RDSSecretRetrievalError(Exception):
    """Exception raised when RDS database credentials cannot be retrieved from Secrets Manager."""


def _set_environment_variable(env_var, value):
    """Helper function to set environment variable if env_var exists."""
    if env_var:
        os.environ[env_var.name] = value


def _get_db_credentials(secrets_client, secret_arn, db_instance_identifier):
    """Retrieve and return database credentials from Secrets Manager."""
    try:
        secret_response = secrets_client.get_secret_value(SecretId=secret_arn)
        return json.loads(secret_response["SecretString"])
    except (
        secrets_client.exceptions.ResourceNotFoundException,
        secrets_client.exceptions.DecryptionFailureException,
        json.JSONDecodeError,
    ) as e:
        # Replaced the placeholder with a custom exception for secret
        # retrieval issues.
        raise RDSSecretRetrievalError(
            f"Error retrieving secret for RDS instance {db_instance_identifier} with ARN {secret_arn}: {e}"
        ) from e


def rds_prep_environment(
    boto_session: BotoSession,
    manifest: InfraManifest,
    annotation: Annotated,
    project_name: str | None,
    # pylint: disable=too-many-locals
):
    infra_associations = inspect_stack(boto_session, manifest, project_name)
    for association in infra_associations:
        if association.annotation == annotation:
            rds_arn = association.rgsa_resource["ResourceARN"]
            db_instance_identifier = rds_arn.split(":")[-1]

            try:
                response = boto_session.session.client("rds").describe_db_instances(
                    DBInstanceIdentifier=db_instance_identifier
                )
                db_instance = response["DBInstances"][0]
            except (
                boto_session.session.client("rds").exceptions.DBInstanceNotFoundFault,
                boto_session.session.client("rds").exceptions.ClientError,
            ) as e:
                # Replaced the placeholder and subsequent error handling with a custom exception.
                raise RDSDescriptionError(
                    f"Error describing RDS instance {db_instance_identifier}: {e}"
                ) from e

            secret_arn = None
            if (
                "MasterUserSecret" in db_instance
                and "SecretArn" in db_instance["MasterUserSecret"]
            ):
                secret_arn = db_instance["MasterUserSecret"]["SecretArn"]
            else:
                secret_arn = "shuttle-db-secret"

            if secret_arn:
                secrets_client = boto_session.session.client("secretsmanager")
                # _get_db_credentials now raises an exception on failure
                secret_data = _get_db_credentials(
                    secrets_client, secret_arn, db_instance_identifier
                )

                if secret_data:
                    endpoint_address = db_instance["Endpoint"]["Address"]
                    endpoint_port = db_instance["Endpoint"]["Port"]

                    _set_environment_variable(
                        annotation.output.engine, secret_data.get("engine", "postgres")
                    )
                    _set_environment_variable(
                        annotation.output.username, secret_data.get("username")
                    )
                    _set_environment_variable(
                        annotation.output.password, secret_data.get("password")
                    )
                    _set_environment_variable(annotation.output.host, endpoint_address)
                    _set_environment_variable(
                        annotation.output.port, str(endpoint_port)
                    )
                    _set_environment_variable(
                        annotation.output.dbname,
                        secret_data.get("dbname", db_instance.get("DBName")),
                    )
                    _set_environment_variable(
                        annotation.output.dbInstanceIdentifier, db_instance_identifier
                    )
            else:
                logger.error(
                    "No MasterUserSecret found for RDS instance %s. "
                    "Cannot retrieve credentials.",
                    db_instance_identifier,
                )


def execute_user_main(module_name: str, module_path: Path):
    """Executes the main function of a user-provided module."""

    logger.info("Running locally...")

    spec = importlib.util.spec_from_file_location(module_name, str(module_path))

    if spec is None:
        raise ImportSpecError(f"Failed to load module spec for {module_name}")

    module = importlib.util.module_from_spec(spec)

    if module is None:
        raise ModuleExecError(f"Failed to create module from spec for {module_name}")

    sys.modules[module_name] = module

    if spec.loader is None:
        raise ModuleExecError("Module loader is None. Cannot execute module.")

    spec.loader.exec_module(module)

    logger.info("Starting local runner...")

    result = asyncio.run(module.main())
    return result



import time
from dataclasses import dataclass
from typing import Any, List

from botocore.exceptions import ClientError, WaiterError

from .common import BotoSession


# Custom exceptions
class CloudFormationError(Exception):
    """Base exception for CloudFormation operations"""


class TemplateValidationError(CloudFormationError):
    """Raised when template validation fails"""


class ChangeSetCreationError(CloudFormationError):
    """Raised when change set creation fails"""


class ChangeSetDescriptionError(CloudFormationError):
    """Raised when describing change set fails"""


class ChangeSetTimeoutError(CloudFormationError):
    """Raised when change set creation times out"""


class StackOperationError(CloudFormationError):
    """Raised when stack create/update operations fail"""


class StackDeletionError(Exception):
    """Raised when stack deletion operation fails"""


class StackDeletionTimeoutError(CloudFormationError):
    """Raised when stack deletion times out"""


class UnparseableChangeSet(Exception):
    """Raised when the ChangeSet cannot be converted to a string"""


class LogsNotFoundError(CloudFormationError):
    """Raised when expected logs or log streams are not found."""


@dataclass
class ChangeSetResult:
    stack_name: str
    change_set_name: str
    change_set_type: str
    stack_id: str
    changes: List[Any]
    describe_response: dict


@dataclass
class StackResourcesResult:
    stack_name: str
    resources: List[Any]
    describe_response: dict


@dataclass
class NoChangeResult:
    operation: str
    message: str
    changes: List[Any]


@dataclass
class ExecutionResult:
    operation: str


class CfClient:
    def __init__(self, boto_session: BotoSession) -> None:
        self.session = boto_session.session
        self.client = self.session.client(
            "cloudformation", endpoint_url=boto_session.endpoint_url
        )
        self.logs_client = self.session.client(
            "logs", endpoint_url=boto_session.endpoint_url
        )

        super().__init__()

    def _stack_exists(self, stack_name: str) -> bool:
        try:
            self.client.describe_stacks(StackName=stack_name)
            return True
        except ClientError as e:
            if "does not exist" in str(e):
                return False
            raise

    def _create_change_set(
        self,
        stack_name: str,
        change_set_name: str,
        change_set_params: dict,
        change_set_type: str,
    ) -> ChangeSetResult | NoChangeResult:
        try:
            # Create the change set
            change_set_response = self.client.create_change_set(**change_set_params)

            # Wait for change set creation to complete
            max_attempts = 60
            attempt = 0

            while attempt < max_attempts:
                try:
                    describe_response = self.client.describe_change_set(
                        StackName=stack_name, ChangeSetName=change_set_name
                    )

                    status = describe_response["Status"]

                    if status == "CREATE_COMPLETE":
                        break
                    if status == "FAILED":
                        # Clean up the failed change set
                        self.client.delete_change_set(
                            StackName=stack_name, ChangeSetName=change_set_name
                        )

                        reason = describe_response.get("StatusReason", "Unknown reason")
                        if (
                            "didn't contain changes" in reason
                            or "No updates are to be performed" in reason
                        ):
                            return NoChangeResult(
                                operation="no_change",
                                message="No updates are to be performed on the stack",
                                changes=[],
                            )

                        raise ChangeSetCreationError(
                            f"Change set creation failed: {reason}"
                        )

                    time.sleep(2)
                    attempt += 1

                except ClientError as e:
                    raise ChangeSetDescriptionError(
                        f"Failed to describe change set: {str(e)}"
                    ) from e

            if attempt >= max_attempts:
                raise ChangeSetTimeoutError("Timeout waiting for change set creation")

            return ChangeSetResult(
                stack_name=stack_name,
                change_set_name=change_set_name,
                change_set_type=change_set_type,
                stack_id=change_set_response["StackId"],
                changes=describe_response.get("Changes", []),
                describe_response=describe_response,
            )

        except ClientError as e:
            raise ChangeSetCreationError(str(e)) from e

    def create_cloudformation_changeset(
        self, stack_name: str, template_json: str
    ) -> ChangeSetResult | NoChangeResult:
        # Validate the template first
        try:
            self.client.validate_template(TemplateBody=template_json)
        except ClientError as e:
            raise TemplateValidationError(
                f"Template validation failed: {str(e)}"
            ) from e

        # Check if stack exists
        exists = self._stack_exists(stack_name)

        # Determine change set type
        change_set_type = "UPDATE" if exists else "CREATE"
        change_set_name = f"{stack_name}-changeset-{int(time.time())}"

        # Create change set parameters
        change_set_params = {
            "StackName": stack_name,
            "ChangeSetName": change_set_name,
            "TemplateBody": template_json,
            "ChangeSetType": change_set_type,
            "Capabilities": [
                "CAPABILITY_IAM",
                "CAPABILITY_NAMED_IAM",
                "CAPABILITY_AUTO_EXPAND",
            ],
        }

        return self._create_change_set(
            stack_name, change_set_name, change_set_params, change_set_type
        )

    def _describe_change_set(
        self,
        stack_name: str,
    ) -> StackResourcesResult:
        try:
            describe_response = self.client.describe_stack_resources(
                StackName=stack_name
            )

            return StackResourcesResult(
                stack_name=stack_name,
                resources=describe_response.get("StackResources", []),
                describe_response=describe_response,
            )

        except ClientError as e:
            raise ChangeSetDescriptionError(
                f"Failed to describe stack resources for {stack_name}: {str(e)}"
            ) from e

    def describe_cloudformation_changeset(
        self, stack_name: str
    ) -> StackResourcesResult | NoChangeResult:
        # Check if stack exists
        if not self._stack_exists(stack_name):
            return NoChangeResult(
                operation="no_stack", message="Stack does not exist", changes=[]
            )

        return self._describe_change_set(stack_name)

    def execute_cloudformation_changeset(
        self, stack_name: str, change_set_name: str, change_set_type: str
    ) -> ExecutionResult:
        try:
            # Execute the change set
            self.client.execute_change_set(
                StackName=stack_name, ChangeSetName=change_set_name
            )

            # Wait for the operation to complete
            try:
                if change_set_type == "CREATE":
                    waiter = self.client.get_waiter("stack_create_complete")
                    waiter.wait(StackName=stack_name)
                else:  # update
                    waiter = self.client.get_waiter("stack_update_complete")
                    waiter.wait(StackName=stack_name)
            except WaiterError as e:
                try:
                    events = self.client.describe_stack_events(StackName=stack_name)
                    create_failure_event = filter(
                        lambda x: x["ResourceStatus"]
                        in [
                            "CREATE_FAILED",
                            "UPDATE_FAILED",
                            "UPDATE_ROLLBACK_IN_PROGRESS",
                        ]
                        and x["ResourceStatusReason"] != "Resource creation cancelled",
                        events["StackEvents"],
                    )
                    reason = list(create_failure_event)[-1]["ResourceStatusReason"]
                except (KeyError, IndexError, ClientError):
                    reason = "unknown"

                raise StackOperationError(
                    f"Stack {change_set_type.lower()} failed: {str(e)}; Reason: {reason}"
                ) from e

            return ExecutionResult(operation=change_set_type.lower())

        except ClientError as e:
            raise StackOperationError(str(e)) from e

    def destroy_stack(self, stack_name: str) -> ExecutionResult:
        try:
            # Check if stack exists, skip if it's already deleted
            if not self._stack_exists(stack_name):
                return ExecutionResult(operation="skip")

            # Delete the stack
            self.client.delete_stack(StackName=stack_name)

            # Wait for deletion to complete
            max_attempts = 360
            attempt = 0

            while attempt < max_attempts:
                describe_response = self.client.describe_stacks(StackName=stack_name)

                stack = describe_response["Stacks"][0]
                status = stack["StackStatus"]

                # Check if the CF stack has failed to delete
                if status == "DELETE_FAILED":
                    raise StackDeletionError(
                        f"Stack {stack_name} failed: {stack['StackStatusReason']}"
                    )

                time.sleep(2)
                attempt += 1

            if attempt >= max_attempts:
                raise StackDeletionTimeoutError("Timeout waiting for deletion")

            return ExecutionResult(operation="delete")

        except ClientError as e:
            # The `describe_stacks` call with throw a missing stack exception
            # if the stack was successfully deleted
            if "does not exist" in str(e):
                return ExecutionResult(operation="delete")
            raise StackOperationError(str(e)) from e

    def get_log_groups_for_stack(self, stack_name: str) -> List[str]:
        try:
            # Get all stack resources
            paginator = self.client.get_paginator("list_stack_resources")
            log_groups = []

            for page in paginator.paginate(StackName=stack_name):
                for resource in page["StackResourceSummaries"]:
                    if resource["ResourceType"] == "AWS::Logs::LogGroup":
                        log_groups.append(resource["PhysicalResourceId"])

            return log_groups
        except ClientError as e:
            raise StackOperationError(
                f"Failed to get log groups for stack {stack_name}: {str(e)}"
            ) from e

    def get_latest_log_stream(self, log_group_name: str) -> str | None:
        response = self.logs_client.describe_log_streams(
            logGroupName=log_group_name,
            orderBy="LastEventTime",
            descending=True,
            limit=1,
        )
        streams = response.get("logStreams", [])
        if streams:
            return streams[0]["logStreamName"]
        return None

    def stream_log_events_from_group(self, log_group_name: str):
        log_stream_name = self.get_latest_log_stream(log_group_name)
        if log_stream_name is None:
            raise LogsNotFoundError(
                f"No log streams found inside log group '{log_group_name}'. It might not exist or the application might not be deployed yet."
            )
        next_token = None

        while True:
            kwargs = {
                "logGroupName": log_group_name,
                "logStreamName": log_stream_name,
                "startFromHead": False,
            }
            if next_token:
                kwargs["nextToken"] = next_token

            response = self.logs_client.get_log_events(**kwargs)
            events = response.get("events", [])
            for event in events:
                yield event["message"], event["timestamp"]

            # Update the token to fetch only new events next time
            new_token = response.get("nextForwardToken")
            if new_token == next_token:
                break
            next_token = new_token


# Take a ChangeSetResult and a list and append string representations of the changes to the list
def change_set_result_str(
    change_set: ChangeSetResult,
    changesets: list[str],
    resource_tag_mapping_list: list[dict] | None = None,
):
    if resource_tag_mapping_list is None:
        resource_tag_mapping_list = []

    if change_set.changes:
        for change in change_set.changes:
            logical_id = change["ResourceChange"]["LogicalResourceId"]
            resource_type = change["ResourceChange"]["ResourceType"]
            action = change["ResourceChange"]["Action"]

            # Initial output
            if action == "Add":
                changesets.append(
                    f"          + {resource_type}\n          id = {logical_id}"
                )
            elif action == "Modify":
                changesets.append(
                    f"          ~ {resource_type}\n          id = {logical_id}"
                )
            elif action == "Remove":
                changesets.append(
                    f"          - {resource_type}\n          id = {logical_id}"
                )
            else:
                raise UnparseableChangeSet(change_set.changes)

            # Add information on resource ARNs if available
            for resource in resource_tag_mapping_list:
                match = {}
                for tag in resource.get("Tags", []):
                    if (
                        tag["Key"] == "aws:cloudformation:stack-id"
                        and tag["Value"] == change_set.stack_id
                    ):
                        match["active"] = True
                    if (
                        tag["Key"] == "aws:cloudformation:logical-id"
                        and tag["Value"] == logical_id
                    ):
                        match["arn"] = resource["ResourceARN"]

                if match.get("active", None) and match.get("arn", None):
                    changesets.append(f"          arn = {match['arn']}")



import json
import logging
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Annotated, Union

import boto3
from aws_cdk import App
from aws_cdk import aws_ecs as ecs
from aws_cdk import aws_iam as iam
from aws_cdk.cx_api import CloudFormationStackArtifact
from botocore.exceptions import BotoCoreError
from pydantic import Field

from shuttle_common import (
    AllowWrite,
    Bucket,
    BucketOutput,
    EcsCronTask,
    EnvVar,
    InfraManifest,
    RdsPostgres,
    RdsPostgresOutput,
)

from . import cf
from .build import DockerClient
from .cdk import EcrRepository, EcsScheduledTaskStack
from .cf import CfClient, ChangeSetResult
from .common import BotoSession

# Configure logging
logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)

logging.getLogger("botocore").setLevel(logging.CRITICAL)


class AWSConnectionError(Exception):
    """Base exception for AWS connection errors"""


class InfraNotProvisionedError(Exception):
    """Raised when the expected infrastructure is not provisioned."""


@dataclass
class InfraAssociation:
    stack_name: str
    changeset_resource: dict
    annotation: Annotated[Union[Bucket, RdsPostgres], Field(..., discriminator="type")]
    rgsa_resource: dict


def _get_aws_config(boto_session: BotoSession):
    """Get AWS configuration details."""
    try:
        account_id = boto_session.session.client(
            "sts", endpoint_url=boto_session.endpoint_url
        ).get_caller_identity()["Account"]
        region = boto_session.session.region_name
        return account_id, region

    except BotoCoreError as e:
        raise AWSConnectionError(e) from e


def _create_resources(
    app: App, repository_name: str, image_uri: str, manifest: InfraManifest
):
    """Create all CDK resources."""
    # Create the ECR repository
    ecr_task = EcrRepository(
        app,
        f"Shuttle-EcrRepository-{repository_name}",
        repository_name=repository_name,
    )
    repository = ecr_task.create()

    # ECS task logic
    ecs_task_result = None
    if isinstance(manifest.service, EcsCronTask):
        # Create the scheduled ECS task
        ecs_task = EcsScheduledTaskStack(
            app,
            f"Shuttle-EcsScheduledTaskStack-{repository_name}",
            cron_schedule=manifest.service.options.schedule,
            repository=repository,
            image_uri=image_uri,
        )
        vpc = ecs_task.create_vpc()
        ecs_task_result = ecs_task.create_service(vpc)
    else:
        raise ValueError(
            f"Unsupported service type: {manifest.service.__class__.__name__}"
        )

    ecs_secrets = {}
    ecs_envvars = {}

    # Resource logic after ECS task is created
    for i, resource in enumerate(manifest.resources):
        if isinstance(resource, Bucket):
            # Define IAM principals to attach to Bucket policy
            s3_principals = [ecs_task_result.task_definition.task_role]
            if resource.options.policies is not None:
                for policy in resource.options.policies:
                    # Check policy type and handle accordingly
                    if isinstance(policy, AllowWrite):
                        s3_principals.append(
                            iam.ArnPrincipal(
                                f"arn:aws:iam::{policy.account_id}:role/{policy.role_name}"
                            )
                        )
            ecs_task.create_bucket(resource.options.bucket_name, s3_principals)
            resource.output = BucketOutput()
        elif isinstance(resource, RdsPostgres):
            database, _ = ecs_task.create_rds(vpc, ecs_task_result.security_group)
            database.secret.grant_read(ecs_task_result.task_definition.execution_role)
            output_kwargs = {}
            for secret_field in (
                "engine",
                "username",
                "password",
                "host",
                "port",
                "dbname",
                "dbInstanceIdentifier",
            ):
                var_name = f"SHUTTLE_RESOURCE_{i:04d}_{secret_field.upper()}"
                ecs_secrets[var_name] = ecs.Secret.from_secrets_manager(
                    database.secret, secret_field
                )
                output_kwargs[secret_field] = EnvVar(name=var_name)
            resource.output = RdsPostgresOutput(**output_kwargs)
        else:
            raise ValueError(
                f"Unsupported resource type: {resource.__class__.__name__}"
            )

    ecs_envvars["__SHUTTLE_RUNTIME_INFRA_OUTPUT"] = manifest.model_dump_json()

    # add env vars and secrets the very last, as they need info from everything else
    for envvar, value in ecs_envvars.items():
        # os.environ[envvar] = value

        ecs_task_result.runtime_container.add_environment(envvar, value)
    for envvar, secret in ecs_secrets.items():
        # os.environ[envvar] = secret
        ecs_task_result.runtime_container.add_secret(envvar, secret)

    return repository, ecs_task_result


def _handle_destroy_operations(session, stacks: list[CloudFormationStackArtifact]):
    """Handle destruction of CloudFormation stacks."""

    cf_client = CfClient(session)

    for stack in reversed(stacks):
        logger.debug("Trying to delete: %s", stack.stack_name)
        cf_client.destroy_stack(stack.stack_name)


def _handle_create_operations(
    session: BotoSession,
    stacks: list[CloudFormationStackArtifact],
    working_directory: Path,
    repository_uri: str,
    image_tag: str,
):
    """Handle creation of CloudFormation stacks."""

    cf_client = CfClient(session)

    # Apply the ECR Repository CF template
    changeset = cf_client.create_cloudformation_changeset(
        stacks[0].stack_name, json.dumps(stacks[0].template)
    )
    changesets = ["src/path/to/task.py", "  └── [~] shuttle_task.cron"]
    cf.change_set_result_str(changeset, changesets)

    if isinstance(changeset, ChangeSetResult):
        cf_client.execute_cloudformation_changeset(
            stacks[0].stack_name, changeset.change_set_name, changeset.change_set_type
        )

    # Build and push the End User image
    docker_client = DockerClient(session)
    docker_client.build(repository_uri, working_directory, image_tag)

    # Create a resource group stagging api client to fetch resource ARNs
    rgsa_client = session.session.client(
        "resourcegroupstaggingapi", endpoint_url=session.endpoint_url
    )

    # Apply remaining stacks
    for stack in stacks[1:]:
        changeset = cf_client.create_cloudformation_changeset(
            stack.stack_name, json.dumps(stack.template)
        )

        if isinstance(changeset, ChangeSetResult):
            cf_client.execute_cloudformation_changeset(
                stack.stack_name, changeset.change_set_name, changeset.change_set_type
            )

        # Fetch resource ARNs for stack resources
        rgsa_response = rgsa_client.get_resources(
            TagFilters=[
                {"Key": "aws:cloudformation:stack-name", "Values": [stack.stack_name]}
            ]
        )

        # Collect a diff changeset for output in the CLI
        cf.change_set_result_str(
            changeset, changesets, rgsa_response["ResourceTagMappingList"]
        )

    logger.info("\n".join(changesets))


def _handle_inspect_operations(
    session: BotoSession,
    stacks: list[CloudFormationStackArtifact],
    manifest: InfraManifest,
) -> list[InfraAssociation]:
    cf_client = CfClient(session)

    # Create a resource group stagging api client to fetch resource ARNs
    rgsa_client = session.session.client(
        "resourcegroupstaggingapi", endpoint_url=session.endpoint_url
    )

    infra_association = []
    for stack in stacks[1:]:
        # Apply the ECR Repository CF template
        changeset = cf_client.describe_cloudformation_changeset(stack.stack_name)

        # Check that the infra is actually provisioned
        if isinstance(changeset, cf.NoChangeResult):
            raise InfraNotProvisionedError(
                f"Infrastructure for stack '{stack.stack_name}' is not provisioned. Please deploy your application first."
            )

        # Fetch resource ARNs for stack resources
        rgsa_response = rgsa_client.get_resources(
            TagFilters=[
                {"Key": "aws:cloudformation:stack-name", "Values": [stack.stack_name]}
            ]
        )

        for annotation in manifest.resources:
            for changeset_resource in changeset.resources:
                # Check that this Annotation is relevant to this Cloudformation Changeset
                if annotation.is_cf_resource(changeset_resource):
                    # Add information on resource ARNs
                    for rgsa_resource in rgsa_response["ResourceTagMappingList"]:
                        match = {}
                        for tag in rgsa_resource.get("Tags", []):
                            if (
                                tag["Key"] == "aws:cloudformation:stack-id"
                                and tag["Value"] == changeset_resource["StackId"]
                            ):
                                match["active"] = True
                            if (
                                tag["Key"] == "aws:cloudformation:logical-id"
                                and tag["Value"]
                                == changeset_resource["LogicalResourceId"]
                            ):
                                match["rgsa"] = rgsa_resource

                        if match.get("active", None) and match.get("rgsa", None):
                            # Collect Annotation with Cloudformation and deployed Resource
                            infra_association.append(
                                InfraAssociation(
                                    stack_name=stack.stack_name,
                                    annotation=annotation,
                                    changeset_resource=changeset_resource,
                                    rgsa_resource=match["rgsa"],
                                )
                            )

    return infra_association


def create_cf_stack(
    session: BotoSession, manifest: InfraManifest, repository_name: str = "simple"
):
    """Main infrastructure deployment function. Constructs the CF stack and client."""
    # Create the CDK App - only one of these can exist
    app = App()
    image_tag = "dev"
    account_id, region = _get_aws_config(session)

    # Handle localstack testing
    if session.endpoint_url and ":" in session.endpoint_url:
        port = session.endpoint_url.split(":")[-1]
        repository_uri = f"{account_id}.dkr.ecr.{region}.localhost.localstack.cloud:{port}/{repository_name}"
    else:
        repository_uri = (
            f"{account_id}.dkr.ecr.{region}.amazonaws.com/{repository_name}"
        )

    image_uri = f"{repository_uri}:{image_tag}"

    # Create all resources
    _create_resources(app, repository_name, image_uri, manifest)

    # Export the CF template, this locks further creation of CDK resources
    stacks = app.synth().stacks

    return stacks, repository_uri, image_tag


def inspect_stack(
    session: BotoSession, manifest: InfraManifest, project_name: str | None
) -> list[InfraAssociation]:
    stacks, _, _ = create_cf_stack(session, manifest, project_name)

    return _handle_inspect_operations(session, stacks, manifest)


def deploy_stack(
    session: BotoSession,
    manifest: InfraManifest,
    working_directory: Path,
    project_name: str | None,
):
    stacks, repository_uri, image_tag = create_cf_stack(session, manifest, project_name)

    # Create the infra resources based on the Cloudformation stack
    _handle_create_operations(
        session, stacks, working_directory, repository_uri, image_tag
    )


def destroy_stack(
    session: boto3.session.Session, manifest: InfraManifest, project_name: str | None
):
    stacks, _, _ = create_cf_stack(session, manifest, project_name)

    _handle_destroy_operations(session, stacks)


def show_logs(session: BotoSession, manifest: InfraManifest, project_name: str | None):
    stacks, _, _ = create_cf_stack(session, manifest, project_name)

    cf_client = CfClient(session)

    for stack in stacks[1:]:
        log_groups = cf_client.get_log_groups_for_stack(stack.stack_name)
        for group in log_groups:
            logger.info("Log group: %s", group)
            for log_line, timestamp in cf_client.stream_log_events_from_group(group):
                t = datetime.fromtimestamp(timestamp / 1000, timezone.utc).strftime(
                    "%Y-%m-%d %H:%M:%S %z"
                )
                logger.info("[%s] %s", t, log_line)



import base64
import io
import json
import logging
import os
import tarfile
from pathlib import Path
from typing import Optional

import docker
import pathspec
from docker.errors import BuildError

from .common import BotoSession

logger = logging.getLogger(__name__)


class DockerContextPathMissing(Exception):
    """Raised when the docker context input directory does not exist"""


class ImagePushError(Exception):
    """Raised when pushing a Docker image to a registry fails"""


# pylint: disable-next=too-few-public-methods
class DockerClient:
    def __init__(self, boto_session: BotoSession) -> None:
        session = boto_session.session
        ecr = session.client("ecr", endpoint_url=boto_session.endpoint_url)

        auth = ecr.get_authorization_token()
        auth_data = auth["authorizationData"][0]
        token = base64.b64decode(auth_data["authorizationToken"]).decode()
        username, password = token.split(":")
        registry = auth_data["proxyEndpoint"].replace("https://", "")

        # Handle local testing
        if registry[-4:] == "4566":
            idx = boto_session.endpoint_url.rindex(":")
            registry = registry[:-4] + boto_session.endpoint_url[idx + 1 :]

        # Initialize Docker client
        self.client = docker.from_env()
        self.client.login(username=username, password=password, registry=registry)

        # Verify Docker is running
        self.client.ping()

    def _load_dockerignore(self, directory: Path) -> Optional[pathspec.PathSpec]:
        """Load .dockerignore file from a directory if it exists."""
        dockerignore_path = directory / ".dockerignore"
        if dockerignore_path.exists():
            try:
                with open(dockerignore_path, "r", encoding="utf-8") as f:
                    patterns = f.read().splitlines()
                return pathspec.PathSpec.from_lines('gitwildmatch', patterns)
            except Exception as e:
                logger.warning("Failed to read .dockerignore file at %s: %s", dockerignore_path, e)
        return None

    def _create_build_context(self, dockerfile_path: Path) -> io.BytesIO:
        """Create a tar archive for the Docker build context."""
        tar_stream = io.BytesIO()

        # Create an in-memory tar file from the build context directory
        logger.debug("Creating build context tar from %s", dockerfile_path)
        with tarfile.open(fileobj=tar_stream, mode="w") as tar:
            base_dir = dockerfile_path.absolute()
            
            for root, dirs, files in base_dir.walk():
                # Load .dockerignore file for current directory
                dockerignore_spec = self._load_dockerignore(Path(root))
                
                # Filter directories using .dockerignore patterns
                if dockerignore_spec:
                    # Remove ignored directories from dirs list to prevent os.walk from traversing them
                    dirs_to_remove = []
                    for dir_name in dirs:
                        dir_rel_path = os.path.relpath(os.path.join(root, dir_name), base_dir)
                        if dockerignore_spec.match_file(dir_rel_path) or dockerignore_spec.match_file(dir_rel_path + '/'):
                            dirs_to_remove.append(dir_name)
                    
                    for dir_name in dirs_to_remove:
                        dirs.remove(dir_name)

                # Process files
                for file in files:
                    file_path = os.path.join(root, file)
                    rel_path = os.path.relpath(file_path, base_dir)
                    
                    # Check if file should be ignored
                    should_ignore = False
                    if dockerignore_spec and dockerignore_spec.match_file(rel_path):
                        should_ignore = True
                    
                    if not should_ignore:
                        tar.add(file_path, arcname=rel_path)

            # Read the contents of the Dockerfile into memory
            dockerfile = Path(__file__).parent.joinpath("Dockerfile").resolve()
            with open(dockerfile, "rb") as fp:
                dockerfile_contents = fp.read()

            # Write the Dockerfile to the tar
            info = tarfile.TarInfo(name="Dockerfile")
            info.size = len(dockerfile_contents)
            tar.addfile(info, io.BytesIO(dockerfile_contents))

        tar_stream.seek(0)
        return tar_stream

    def _push_image(self, repository_name: str, image_tag: str) -> None:
        """Helper to push Docker image and handle push logs/errors."""
        full_tag = f"{repository_name}:{image_tag}"
        logger.debug("Pushing image to repository: %s", full_tag)
        push_errors = []

        for log_line in self.client.images.push(
            repository_name, image_tag
        ).splitlines():
            log = json.loads(log_line)
            if "stream" in log and log["stream"].strip():
                logger.debug(log["stream"].strip())
            elif "status" in log and log["status"].strip():
                logger.debug(log["status"].strip())

            # Consolidate error checking for 'error' and 'errorDetail' keys
            if "error" in log or ("errorDetail" in log and log["errorDetail"]):
                error_msg = log.get("error", "").strip()
                error_detail_msg = log.get("errorDetail", {}).get("message", "").strip()

                full_error = "Unknown Docker image push error"
                if error_msg and error_detail_msg:
                    full_error = f"{error_msg} ({error_detail_msg})"
                elif error_msg:
                    full_error = error_msg
                elif error_detail_msg:
                    full_error = error_detail_msg

                push_errors.append(full_error)
                logger.error("Docker image push error: %s", full_error)

        if push_errors:
            raise ImagePushError(
                f"Failed to push Docker image: {' | '.join(push_errors)}"
            )

        logger.debug("Successfully pushed image: %s", full_tag)

    def build(
        self,
        repository_name: str,
        dockerfile_path: Path,
        image_tag: str = "latest",
    ) -> str:
        if not dockerfile_path.exists():
            raise DockerContextPathMissing(
                "Docker context directory does not exist or is not a directory"
            )

        try:
            image_name = f"{repository_name}:{image_tag}"
            logger.debug("Building Docker image with context from %s", dockerfile_path)

            tar_stream = self._create_build_context(dockerfile_path)

            logger.debug("Building image: %s", image_name)

            # Build the Docker image using the custom context
            _image, build_logs = self.client.images.build(
                fileobj=tar_stream,
                custom_context=True,
                tag=image_name,
                rm=True,
                forcerm=True,
            )

            # Log build output
            for log in build_logs:
                if "stream" in log:
                    log_line = log["stream"].strip()
                    if log_line:
                        logger.debug(log_line)

            logger.debug("Successfully built image: %s", image_name)

            # Push the image
            self._push_image(repository_name, image_tag)

            return image_name

        except BuildError as e:
            for log in getattr(e, "build_log", []):
                if "stream" in log:
                    log_line = log["stream"].strip()
                    if log_line:
                        logger.error(log_line)
            raise



import logging
import os
from pathlib import Path

os.environ["JSII_SILENCE_WARNING_UNTESTED_NODE_VERSION"] = "1"

import boto3
import click
from docker.errors import DockerException

from shuttle_common import RdsPostgres
from shuttle_infra import (
    AWSConnectionError,
    deploy_stack,
    destroy_stack,
    show_logs,
    InfraNotProvisionedError,
)
from shuttle_infra.build import DockerContextPathMissing, ImagePushError
from shuttle_infra.cf import (
    CloudFormationError,
    StackDeletionError,
    StackDeletionTimeoutError,
    UnparseableChangeSet,
    LogsNotFoundError,
    StackOperationError
)
from shuttle_infra.common import BotoSession

from .lib import InfraDiscoveryError, discover_project_and_infra
from .local import (
    ImportSpecError,
    ModuleExecError,
    RDSDescriptionError,
    RDSSecretRetrievalError,
    execute_user_main,
    rds_prep_environment,
)


# Configure basic logging setup if not already configured
if not logging.getLogger().handlers:
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )


def configure_debug_logging(ctx, param, value):
    """Callback to enable debug logging for shuttle modules."""
    if value:
        logging.getLogger("shuttle_common").setLevel(logging.DEBUG)
        logging.getLogger("shuttle_infra").setLevel(logging.DEBUG)
        logging.getLogger(__name__).setLevel(logging.DEBUG)
        click.echo("Debug logging enabled for shuttle-related modules.")
    return value


# Define the common debug option
debug_option = click.option(
    "--debug",
    is_flag=True,
    help="Enable debug logging for shuttle based modules.",
    callback=configure_debug_logging,
    expose_value=True,
    is_eager=True,
)


@click.group()
def app():
    """Shuttle CLI for Python 🤨🐍"""
    pass


@app.command()
@debug_option
@click.argument("path", required=False, default=".")
def deploy(path, debug):
    """Provision, build, and deploy the application."""

    session = boto3.session.Session()
    boto_session = BotoSession(endpoint_url=None, session=session)
    try:
        project_root, manifest, project_name = discover_project_and_infra(path)
    except FileNotFoundError as e:
        click.echo(
            f"Required file or directory not found: {e}. Please ensure the project structure is valid and all necessary files exist.",
            err=True,
        )
        return
    except InfraDiscoveryError as e:
        click.echo(f"Error discovering project infrastructure: {e}", err=True)
        return

    try:
        deploy_stack(boto_session, manifest, project_root, project_name)
    except UnparseableChangeSet as e:
        click.echo(
            f"Failed to parse CloudFormation change set due to an unknown or unexpected change type: {e}. This indicates an internal issue or an unexpected AWS CloudFormation response. Please report this issue.",
            err=True,
        )
    except CloudFormationError as e:  # Catch all CloudFormation-related errors (TemplateValidationError, ChangeSetCreationError, ChangeSetDescriptionError, ChangeSetTimeoutError, StackOperationError)
        click.echo(f"CloudFormation operation failed: {e}", err=True)
    except AWSConnectionError as e:
        click.echo(f"Failed to connect to AWS: {e}", err=True)
    except DockerException as e:
        click.echo(f"Fatal Docker error: {e}", err=True)
    except ImagePushError as e:
        click.echo(f"Failed to push Docker image: {e}", err=True)
    except DockerContextPathMissing as e:
        click.echo(
            f"Docker build context path missing: {e}. Please ensure the project directory exists and is accessible.",
            err=True,
        )


@app.command()
@debug_option
@click.argument("path", required=False, default=".")
def destroy(path, debug):
    """Destroy the stack."""

    session = boto3.session.Session()
    boto_session = BotoSession(endpoint_url=None, session=session)
    try:
        _, manifest, project_name = discover_project_and_infra(path)
    except FileNotFoundError as e:
        click.echo(
            f"Required file or directory not found: {e}. Please ensure the project structure is valid and all necessary files exist.",
            err=True,
        )
        return
    except InfraDiscoveryError as e:
        click.echo(f"Error discovering project infrastructure: {e}", err=True)
        return

    try:
        destroy_stack(boto_session, manifest, project_name)
    except StackDeletionTimeoutError as e:
        click.echo(
            f"Deletion of the CloudFormation stack timed out: {e}. Please check your AWS console for the stack's status and consider manual intervention if necessary.",
            err=True,
        )
    except StackDeletionError as e:
        click.echo(f"Deprovisioning of Cloudformation resources failed: {e}", err=True)
    except CloudFormationError as e:  # Catch all CloudFormation-related errors (StackOperationError, StackDeletionTimeoutError)
        click.echo(f"CloudFormation operation failed: {e}", err=True)
    except AWSConnectionError as e:
        click.echo(f"Failed to connect to AWS: {e}", err=True)


@app.command()
@debug_option
@click.argument("path", required=False, default=".")
def logs(path, debug):
    """Show logs from the deployed app."""

    session = boto3.session.Session()
    boto_session = BotoSession(endpoint_url=None, session=session)
    try:
        _, manifest, project_name = discover_project_and_infra(path)
    except FileNotFoundError as e:
        click.echo(
            f"Required file or directory not found: {e}. Please ensure the project structure is valid and all necessary files exist.",
            err=True,
        )
        return
    except InfraDiscoveryError as e:
        click.echo(f"Error discovering project infrastructure: {e}", err=True)
        return

    try:
        show_logs(boto_session, manifest, project_name)
    except LogsNotFoundError as e:
        click.echo(f"Logs not found: {e}", err=True)
    except StackOperationError as e:
        click.echo(f"Failed to retrieve logs due to a stack operation error: {e}", err=True)
    except CloudFormationError as e: # Catch other CloudFormation-related errors (e.g., TemplateValidationError, ChangeSetCreationError, etc. from get_log_groups_for_stack)
        click.echo(f"CloudFormation operation failed: {e}", err=True)
    except AWSConnectionError as e:
        click.echo(f"Failed to connect to AWS: {e}", err=True)


@app.command()
@debug_option
@click.argument("path", required=False, default=".")
def run(path, debug):
    rpath = Path(path).resolve()
    if (entrypoint := rpath.joinpath("main.py")).exists():
        pass
    elif (entrypoint := rpath.joinpath("__main__.py")).exists():
        pass
    else:
        click.echo(f"Error: didn't find a main python file in {rpath}", err=True)
        return

    session = boto3.session.Session()
    boto_session = BotoSession(endpoint_url=None, session=session)
    try:
        _, manifest, project_name = discover_project_and_infra(path)
    except FileNotFoundError as e:
        click.echo(
            f"Required file or directory not found: {e}. Please ensure the project structure is valid and all necessary files exist.",
            err=True,
        )
        return
    except InfraDiscoveryError as e:
        click.echo(f"Error discovering project infrastructure: {e}", err=True)
        return

    try:
        # We need to specifically handle Rds, because CDK doesn't populate
        # connection details about Posgres until it applies infra - so if we're
        # just running locally we need to fetch RDS secrets manually.
        for annotation in manifest.resources:
            if isinstance(annotation, RdsPostgres):
                rds_prep_environment(boto_session, manifest, annotation, project_name)

        if debug:
            click.echo("Manifest Model Dump:")
            click.echo(manifest.model_dump())
        os.environ["__SHUTTLE_RUNTIME_INFRA_OUTPUT"] = manifest.model_dump_json()

        execute_user_main(path, entrypoint)
    except InfraNotProvisionedError as e:
        click.echo(f"Error: {e}", err=True)
    except CloudFormationError as e:  # Catch CloudFormation errors (e.g., ChangeSetDescriptionError from rds_prep_environment)
        click.echo(f"CloudFormation operation failed: {e}", err=True)
    except RDSDescriptionError as e:
        click.echo(f"Failed to describe RDS instance: {e}", err=True)
    except RDSSecretRetrievalError as e:
        click.echo(f"Failed to retrieve RDS secrets: {e}", err=True)
    except ImportSpecError as e:
        click.echo(f"Failed to load user module: {e}", err=True)
    except ModuleExecError as e:
        click.echo(f"Failed to execute user module: {e}", err=True)
    except AttributeError as e:
        click.echo(
            f"The application's 'main' function was not found or is not callable: {e}",
            err=True,
        )
    except Exception as e:
        click.echo(
            f"An unexpected error occurred during local execution: {e}", err=True
        )


if __name__ == "__main__":
    app()



import functools
import inspect
import logging
import os
import typing

from shuttle_common import (
    InfraManifest,
    EcsCronTask,
    EcsCronTaskOptions,
    AllResourcesUnion,
)

logger = logging.getLogger(__name__)


def cron(*args, **kwargs):
    logger.debug(f"decorator called with {args=} {kwargs=}")

    def wrapper(func):
        anno = inspect.get_annotations(func)
        types = typing.get_type_hints(func)
        logger.debug(f"in wrapper {func=} {anno=} {types=}")

        # transform annotations into instances of classes with the args provided
        resources: list[AllResourcesUnion] = []
        for _arg_ident, annotated in anno.items():
            logger.debug(f"{_arg_ident!r} annotations:")
            for meta in annotated.__metadata__:
                logger.debug("-", meta)

            # construct the class in the annotation with the first annotation (options) if there is one
            options = annotated.__metadata__[0] if annotated.__metadata__ else None
            resource = annotated(options=options)
            resources.append(resource)

        infra = InfraManifest(
            service=EcsCronTask(options=EcsCronTaskOptions(**kwargs)),
            resources=resources,
        )

        @functools.wraps(func)
        async def wrapped():
            logger.debug("--- initializing resources ---")
            # replace the "requested" infra manifest with the same manifest populated in the env with the outputs
            infra = InfraManifest.model_validate_json(
                os.environ.get("__SHUTTLE_RUNTIME_INFRA_OUTPUT")
            )
            resources = infra.resources
            # initialize each main function resource and pass them in
            for arg in resources:
                arg.init()
            logger.debug("--- before main executes ---")
            result = await func(*resources)
            logger.debug("--- after main executes ---")
            return result

        wrapped.__shuttle_infra__ = infra

        return wrapped

    return wrapper



import asyncio
import os
import sys

from shuttle_common import InfraManifest

VERSION = "0.1.0-beta"
VERSION_STRING = f"shuttle_runtime (Python) {VERSION}"


# takes in the async def instead of a coroutine so that we can access
# the __shuttle_infra__ that the decorator attaches to it
def main(user_fnc):
    # import warnings
    # # ignore warning since we might not await the coroutine
    # warnings.filterwarnings("ignore", category=RuntimeWarning, message=".*coroutine '.*' was never awaited.*")

    for arg in sys.argv:
        if arg == "--version":
            print(VERSION_STRING)
            return

    if os.environ.get("SHUTTLE_RUNTIME") is None:
        print("Use 'shuttle run' to run a function wrapped in shuttle_runtime.main()")
        return

    emit_infra_manifest = os.environ.get("SHUTTLE_GET_INFRA_MANIFEST")
    if emit_infra_manifest == "true":
        if hasattr(user_fnc, "__shuttle_infra__") and isinstance(
            user_fnc.__shuttle_infra__, InfraManifest
        ):
            print(user_fnc.__shuttle_infra__.model_dump_json(), end="")
    else:
        print(f"{VERSION_STRING} starting")
        asyncio.run(user_fnc())



import logging
import subprocess
import tomllib
from pathlib import Path

from shuttle_common import InfraManifest
from shuttle_infra import build

logger = logging.getLogger(__name__)


class InfraDiscoveryError(Exception):
    """Custom exception for errors during infrastructure discovery."""


def discover_project_and_infra(path_str: str):
    path = Path(path_str).resolve()
    project_name = None

    # Find a venv in project folder or closest parent
    for parent in [path, *path.parents]:
        venv_dirs = [".venv", "venv"]
        for v in venv_dirs:
            if (venv := parent.joinpath(v)).exists():
                break
        else:
            continue
        break
    else:
        raise FileNotFoundError("No venv found in project or parents")
    logging.debug("Found venv in %s", parent)

    # If we're running the user-project outside of shuttle-cobra,
    # this should resolve to the Dockerfile inside of shuttle-cobra
    if not (_ := Path(build.__file__).parent.joinpath("Dockerfile")).exists():
        raise FileNotFoundError("No Dockerfile found in project root")
    logging.debug("Found Dockerfile in %s", Path(build.__file__).parent)

    if (entrypoint := path.joinpath("main.py")).exists():
        program = [
            "/usr/bin/env",
            "bash",
            "-c",
            f"source {venv}/bin/activate; exec python3 {entrypoint}",
        ]
    elif path.joinpath("__main__.py").exists():
        program = [
            "/usr/bin/env",
            "bash",
            "-c",
            f"source {venv}/bin/activate; exec python3 -m {path.relative_to(Path.cwd())}",
        ]
    else:
        raise FileNotFoundError(f"didn't find a main python file in {path}")

    if (pyproject := parent.joinpath("pyproject.toml")).exists():
        with open(pyproject, "rb") as f:
            data = tomllib.load(f)

        project_name = data.get("project", {}).get("name")

    proc = subprocess.run(
        program,
        env={"SHUTTLE_RUNTIME": "true", "SHUTTLE_GET_INFRA_MANIFEST": "true"},
        capture_output=True,
        check=False,
    )
    try:
        proc.check_returncode()
        manifest = InfraManifest.model_validate_json(proc.stdout)
    except Exception as e:
        stdout_str = proc.stdout.decode("utf-8", errors="ignore").strip()
        stderr_str = proc.stderr.decode("utf-8", errors="ignore").strip()
        error_details = f"Original error: {e}"
        if stdout_str:
            error_details += f"\nProcess stdout:\n{stdout_str}"
        if stderr_str:
            error_details += f"\nProcess stderr:\n{stderr_str}"
        raise InfraDiscoveryError(
            f"Failed to extract infrastructure definitions from project. {error_details}"
        ) from e

    logging.debug("infra manifest output: %s", manifest)

    return parent, manifest, project_name



from dataclasses import dataclass
from typing import Optional
from aws_cdk import (
    App,
    Duration,
    Stack,
    RemovalPolicy,
    aws_s3 as s3,
    aws_ecs as ecs,
    aws_ec2 as ec2,
    aws_events as events,
    aws_events_targets as targets,
    aws_ecr as ecr,
    aws_iam as iam,
    aws_rds as rds,
    LegacyStackSynthesizer,
)


@dataclass
class RdsConfig:
    # pylint: disable=too-many-instance-attributes
    database_name: str = "shuttle_db"
    username: str = "shuttle"
    engine_version: rds.PostgresEngineVersion = rds.PostgresEngineVersion.VER_17_4
    instance_type: ec2.InstanceType = ec2.InstanceType.of(
        ec2.InstanceClass.BURSTABLE3, ec2.InstanceSize.MICRO
    )
    allocated_storage: int = 20
    multi_az: bool = False
    backup_retention_days: Duration = Duration.days(7)
    deletion_protection: bool = False
    secret_name: str = "shuttle-db-secret"


@dataclass
class EcsScheduledTaskResult:
    task_definition: ecs.FargateTaskDefinition
    rule: events.Rule
    cluster: ecs.Cluster
    runtime_container: ecs.ContainerDefinition
    security_group: ec2.SecurityGroup


class EcrRepository(Stack):
    def __init__(self, scope: App, construct_id: str, **kwargs) -> None:
        self.repository_name = kwargs.pop("repository_name")

        kwargs["synthesizer"] = LegacyStackSynthesizer()

        super().__init__(scope, construct_id, **kwargs)

    @property
    def name(self) -> str:
        return self.repository_name

    def create(self):
        # Create a private ECR repository
        return ecr.Repository(
            self,
            "ShuttleScheduledTaskRepo",
            repository_name=self.repository_name,
            image_scan_on_push=True,
            removal_policy=RemovalPolicy.DESTROY,  # For easy cleanup during development
            empty_on_delete=True,  # Allows a CF destroy to destroy images that were pushed
        )


class EcsScheduledTaskStack(Stack):
    def __init__(self, scope: App, construct_id: str, **kwargs) -> None:
        self.cron_schedule = kwargs.pop("cron_schedule", "0 12 * * ? *")
        self.repository = kwargs.pop("repository")
        self.image_uri = kwargs.pop("image_uri")

        kwargs["synthesizer"] = LegacyStackSynthesizer()

        super().__init__(scope, construct_id, **kwargs)

    def create_vpc(self) -> ec2.Vpc:
        vpc = ec2.Vpc(self, "ShuttleVpc", max_azs=2)

        return vpc

    def create_service(self, vpc: ec2.Vpc) -> EcsScheduledTaskResult:
        # Create an ECS cluster
        cluster = ecs.Cluster(self, "ShuttleCluster", vpc=vpc)

        # Create a Fargate task definition
        task_definition = ecs.FargateTaskDefinition(
            self,
            "ShuttleScheduledTaskDefinition",
            memory_limit_mib=512,
            cpu=256,
        )

        # Add the required ECR permissions to the execution role
        task_definition.add_to_execution_role_policy(
            iam.PolicyStatement(actions=["ecr:getAuthorizationToken"], resources=["*"])
        )

        repository_arn = self.repository.repository_arn

        task_definition.add_to_execution_role_policy(
            iam.PolicyStatement(
                actions=["ecr:BatchGetImage", "ecr:GetDownloadUrlForLayer"],
                resources=[repository_arn],
            )
        )

        # Add container to the task definition
        runtime_container = task_definition.add_container(
            "Runtime",
            image=ecs.ContainerImage.from_registry(self.image_uri),
            logging=ecs.LogDrivers.aws_logs(stream_prefix="scheduled-task"),
        )

        # Create the EventBridge rule with the provided cron schedule
        rule = events.Rule(
            self,
            "ScheduleRule",
            schedule=events.Schedule.expression(f"cron({self.cron_schedule})"),
        )

        security_group = ec2.SecurityGroup(
            self,
            "ShuttleEcsSecurityGroup",
            vpc=vpc,
            description="Security group for Shuttle ECS task",
            allow_all_outbound=True,
        )

        # Add the ECS task as a target for the rule
        rule.add_target(
            targets.EcsTask(
                cluster=cluster,
                task_definition=task_definition,
                subnet_selection=ec2.SubnetSelection(subnet_type=ec2.SubnetType.PUBLIC),
                security_groups=[security_group],
            )
        )

        return EcsScheduledTaskResult(
            task_definition=task_definition,
            rule=rule,
            cluster=cluster,
            runtime_container=runtime_container,
            security_group=security_group,
        )

    def create_bucket(
        self, bucket_name: str, additional_principals: list[iam.IPrincipal]
    ) -> s3.Bucket:
        # Create S3 bucket with security best practices
        bucket = s3.Bucket(
            self,
            "ShuttleS3Bucket",
            bucket_name=bucket_name,
            object_ownership=s3.ObjectOwnership.BUCKET_OWNER_ENFORCED,
            public_read_access=False,
            block_public_access=s3.BlockPublicAccess.BLOCK_ALL,
            versioned=True,
            encryption=s3.BucketEncryption.S3_MANAGED,
            removal_policy=RemovalPolicy.DESTROY,  # For easy cleanup during development
        )

        for role in additional_principals:
            bucket.add_to_resource_policy(
                iam.PolicyStatement(
                    effect=iam.Effect.ALLOW,
                    principals=[role],
                    actions=["s3:ListBucket"],
                    resources=[bucket.bucket_arn],
                )
            )

            bucket.add_to_resource_policy(
                iam.PolicyStatement(
                    effect=iam.Effect.ALLOW,
                    principals=[role],
                    actions=["s3:GetObject", "s3:PutObject", "s3:DeleteObject"],
                    resources=[bucket.arn_for_objects("*")],
                )
            )

        return bucket

    def create_rds(
        self,
        vpc: ec2.Vpc,
        ecs_security_group: ec2.SecurityGroup,
        config: Optional[RdsConfig] = None,
    ):
        if config is None:
            config = RdsConfig()

        # Create a subnet group for the RDS instance
        subnet_group = rds.SubnetGroup(
            self,
            "ShuttleDbSubnetGroup",
            description="Subnet group for Shuttle RDS instance",
            vpc=vpc,
            vpc_subnets=ec2.SubnetSelection(subnet_type=ec2.SubnetType.PUBLIC),
        )

        # Create a security group for the RDS instance
        security_group = ec2.SecurityGroup(
            self,
            "ShuttleDbSecurityGroup",
            vpc=vpc,
            description="Security group for Shuttle RDS instance",
            allow_all_outbound=False,
        )
        security_group.add_ingress_rule(
            peer=ec2.Peer.any_ipv4(),
            connection=ec2.Port.tcp(5432),
            description="Allow public access to RDS on PSQL port",
        )

        credentials = rds.Credentials.from_generated_secret(
            config.username,
            secret_name=config.secret_name,
        )

        # Create the RDS instance
        db_instance = rds.DatabaseInstance(
            self,
            "ShuttleDatabase",
            engine=rds.DatabaseInstanceEngine.postgres(version=config.engine_version),
            instance_type=config.instance_type,
            vpc=vpc,
            subnet_group=subnet_group,
            security_groups=[security_group],
            database_name=config.database_name,
            credentials=credentials,
            allocated_storage=config.allocated_storage,
            multi_az=config.multi_az,
            backup_retention=config.backup_retention_days,
            deletion_protection=config.deletion_protection,
            auto_minor_version_upgrade=True,
            delete_automated_backups=True,
            removal_policy=RemovalPolicy.DESTROY,  # For easy cleanup during development
            publicly_accessible=True,  # Mark the database as publicly accessible
        )

        return db_instance, config
